import numpy as np
from matplotlib import pyplot as plt


def visualize_check(img, title):
  # 显示一个带确认按钮的窗口，如果确认则返回True，否则返回False
  pass


def show_img(img):
  # TODO: img valid test. use a gui to validate.
  is_show = 1
  if is_show:
    plt.imshow(img)
    plt.show()


def gen_img(size=(35, 25, 3)):
  return np.random.randint(0, 255, size=size, dtype=np.uint8)



# generate dataset path.
import os
import pathlib

import streamlit as st

from meterviewer.config import get_root_path


def main():
  st.title("Generate dataset path")
  p = get_root_path()
  p = pathlib.Path(st.text_input("root path:", value=p))
  dataset_name = st.selectbox("dataset name", options=os.listdir(p))

  ds_path = p / dataset_name
  st.button("finish")
  st.text(ds_path)


main()



# read dataset from file system.
import typing as t
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from pathlib import Path as P

from meterviewer import func as F
from meterviewer import types as T

from .types import typeOfrect


@dataclass
class RectO(object):
  """矩形对象类，用于存储和处理矩形框的坐标信息

  属性:
      xmin (str): 矩形左边界x坐标
      ymin (str): 矩形上边界y坐标
      xmax (str): 矩形右边界x坐标
      ymax (str): 矩形下边界y坐标
  """

  xmin: str = ""
  ymin: str = ""
  xmax: str = ""
  ymax: str = ""

  def check(self):
    # assert rect.xmin != "" and rect.ymin != "" and rect.xmax != "" and rect.ymax != "", rect
    assert self.xmin != "" and self.ymin != "" and self.xmax != "" and self.ymax != "", self

  def to_dict(self) -> T.Rect:
    return T.Rect.model_validate(
      {
        "xmin": self.xmin,
        "ymin": self.ymin,
        "xmax": self.xmax,
        "ymax": self.ymax,
      }
    )

  def __str__(self):
    return f"RectO({self.xmin}, {self.ymin}, {self.xmax}, {self.ymax})"

  def __repr__(self):
    return self.__str__()


def read_rect_from_node(root: t.Iterable) -> t.Tuple[str, T.Rect]:
  """从XML节点中读取矩形框信息

  Args:
      root (Iterable): XML根节点

  Returns:
      Tuple[str, T.Rect]: 返回值和矩形框字典的元组
  """
  val, rect_dict = "", RectO()
  for child in root:
    # find object
    if not child.tag == "object":
      continue
    for subchild in child:
      if subchild.tag == "name":
        val = F.will_be_str(subchild.text)

      # print(subchild.tag, subchild.text)
      if subchild.tag == "bndbox":
        for sub in subchild:
          setattr(rect_dict, sub.tag, sub.text)

  return val, rect_dict.to_dict()


def read_xml(filename: P):
  """读取 xml 文件，获取 root 节点"""
  assert filename.exists(), f"{filename}文件不存在"
  tree = ET.parse(filename)
  return tree.getroot()


def read_xml_to_get(
  filename: P,
  read_func: t.Callable[[t.Any], t.Any] = read_rect_from_node,
):
  """读取并解析XML文件

  Args:
      filename (P): XML文件路径
      read_func (Callable): 用于处理XML根节点的回调函数

  Returns:
      Any: 回调函数的返回结果
  """
  assert filename.exists(), f"{filename}文件不存在"
  return read_func(read_xml(filename))


def read_single_digit_rect(filename) -> t.List[RectO]:
  """读取单个数字的矩形框信息

  Args:
      filename: XML文件路径

  Returns:
      List[RectO]: 包含6个数字位置的矩形框列表
  """

  def func(root: t.Iterable) -> t.List[RectO]:
    def find_no(node) -> t.Tuple[int, RectO]:
      no = -1
      rect = RectO()
      for subchild in node:
        if subchild.tag == "no":
          no = int(subchild.text.strip())
        else:
          rect = set_rect(rect, subchild)
      rect.check()
      assert no != -1, "cannot find no number"
      return no, rect

    def set_rect(rect: RectO, sub) -> RectO:
      assert hasattr(rect, sub.tag), (sub.tag, sub.text)
      setattr(rect, sub.tag, sub.text)
      assert getattr(rect, sub.tag) != ""
      return rect

    def num_check():
      seta = {0, 1, 2, 3, 4, 5}
      setb = set()

      def is_valid():
        return seta == setb, (seta, setb)

      def set_num(no):
        setb.add(no)

      return set_num, is_valid

    digit_rect = [RectO() for _ in range(6)]
    set_num, is_valid = num_check()
    is_loop, set_loop = F.looped()

    for child in root:
      if child.tag == "digit":
        no, rect = find_no(child)
        _ = set_loop(), set_num(no)
        digit_rect[no] = rect

    assert is_loop(), "node has no child"
    cond, _ = is_valid()
    assert cond

    # return digit_rect, root, find_no, find_rect
    return digit_rect

  return read_xml_to_get(filename, func)


def get_single_digit_values(filename: P) -> t.Tuple[str, T.Rect]:
  """获取单个数字的值和位置信息

  Args:
      filename (P): XML文件路径

  Returns:
      Tuple[str, T.Rect]: 返回值和矩形框信息的元组
  """
  val, _ = read_xml_to_get(filename, read_rect_from_node)
  block_pos = read_xml_to_get(filename, read_single_digit_rect)
  return val, block_pos


def read_rect_from_file(xml_path: P, type_: typeOfrect):
  """从文件中读取矩形框信息

  Args:
      xml_path (P): XML文件路径
      type_ (typeOfrect): 矩形框类型，可以是'single'或'block'

  Returns:
      Any: 根据类型返回相应的矩形框信息
  """
  assert xml_path.suffix == ".xml"
  function_map: t.Mapping[typeOfrect, t.Callable[[P], t.Any]] = {
    "single": read_single_digit_rect,
    "block": get_rectangle,
  }
  func = function_map[type_]
  return func(xml_path)


def get_rectangle(filename: P) -> T.Rect:
  """获取矩形框信息

  Args:
      filename (P): XML文件路径

  Returns:
      T.Rect: 矩形框字典
  """
  _, rect = read_xml_to_get(filename, read_rect_from_node)
  return rect


def get_xml_config_path(img_path: P, types: t.Literal["value", "block", "single"] = "value") -> P:
  """根据图片路径获取对应的XML配置文件路径

  Args:
      img_path (P): 图片文件路径
      types (Literal): 配置类型，可以是'value'、'block'或'single'

  Returns:
      P: XML配置文件路径
  """
  dataset_path = img_path.parent

  def value_path():
    config_p = P(dataset_path) / "baocun"
    assert img_path.suffix in (".jpg", ".jpeg")
    filename = img_path.stem + ".xml"
    return config_p / filename

  def block_path():
    return P(dataset_path) / "config" / "block.xml"

  def single_path():
    return P(dataset_path) / "config" / "res.xml"

  funcs: t.Mapping[str, t.Callable] = {
    "value": value_path,
    "block": block_path,
    "single": single_path,
  }

  return funcs[types]()


def get_xml_config(img_path: P) -> t.Tuple[str, T.Rect]:
  """获取XML配置信息

  Args:
      img_path (P): 图片文件路径

  Returns:
      Tuple[str, T.Rect]: 返回值和矩形框信息的元组
  """
  return read_xml_to_get(
    filename=get_xml_config_path(img_path),
    read_func=read_rect_from_node,
  )



from pathlib import Path as P


def check_if_meterdataset(path: P) -> bool:
  """
  Check if the dataset is meter-formatted.

  """

  must_exist_folder = [
    "baocun",
    "config",
    "coor_all_img_np",
    "Digit",
    "id",
    "ImageSets",
    "ImageSets_block_zoom",
    "ImageSets_seg",
    "Numpy_block_zoom",
    "Numpy_seg",
  ]

  must_exist_folder = [P(folder) for folder in must_exist_folder]

  for folder in must_exist_folder:
    if not (path / folder).exists():
      raise ValueError(f"Folder {folder} does not exist in {path}")



import typing as t

import numpy as np
from pydantic import BaseModel

# Numpy typed Image
NpImage = np.ndarray

Label = np.ndarray
ImgSize = t.Union[t.List[int], t.Tuple[int, int]]

ImgDataset = np.ndarray
LabelData = np.ndarray


Func = t.Callable

ImgList = t.List[NpImage]
DigitStr = t.List[str]
DigitInt = t.List[int]

CheckFunc = t.Callable[[t.Any], t.Any]
JoinFunc = t.Callable[[ImgList, CheckFunc], NpImage]
NameFunc = t.Callable[[], t.Tuple[str, str]]

x_name: str = "x_train.npy"
y_name: str = "y_train.npy"
x_test: str = "x_test.npy"
y_test: str = "y_test.npy"


def isImgDataset(x: t.Any) -> bool:
  return isinstance(x, np.ndarray) and len(x.shape) == 4


def isLabelData(y: t.Any) -> bool:
  return isinstance(y, np.ndarray) and len(y.shape) == 2


class Rect(BaseModel):
  xmin: int
  ymin: int
  xmax: int
  ymax: int



import functools
import typing as t

from loguru import logger

from meterviewer import T
from meterviewer.img import process


def join_with_fix(
  imglist: T.ImgList,
  check_func: t.Callable,
  fix_func: t.Callable,
) -> T.NpImage:
  """修饰 join_func"""
  # merge images horizontally
  try:
    return process.join_img(imglist, check_func)
  except ValueError as e:
    logger.debug(e)
    imglist = fix_func(imglist)
  return process.join_img(imglist, check_func)


join_with_resize: T.JoinFunc = functools.partial(
  join_with_fix, fix_func=process.resize_imglist
)



from pydantic import BaseModel


class Item(BaseModel):
  filepath: str
  dataset: str

  xmin: float
  xmax: float
  ymin: float
  ymax: float


class MeterDB(BaseModel):
  """the data load from the meterdb.json"""

  data: list[Item]



import typing as t
import numpy as np


class Validator(object):
  # a image validator

  def is_img(self, img: np.ndarray) -> bool:
    conds = [
      lambda x: isinstance(x, np.ndarray),
      lambda x: len(x.shape) == 3,  # gray and color.
      lambda x: x.dtype == np.uint8,
      lambda x: x.shape[2] in [1, 3],  # gray and color.
      lambda x: x.shape[0] > 0 and x.shape[1] > 0,
    ]
    for c in conds:
      if not c(img):
        return False
    return True

  def valid_shape(self, img: np.ndarray, shape: t.Tuple) -> bool:
    return img.shape == shape



import typing as t
from meterviewer import files
from meterviewer.datasets import dataset
import pathlib
from .types import viewReturn


def from_filesystem(root_path: pathlib.Path) -> viewReturn:
  for dataset_name in dataset.get_dataset_list(root_path):
    img_p = files.scan_pics(dataset.get_dataset_path(root_path, str(dataset_name)))
    img_p = list(img_p)
    for p in img_p:
      yield p, dataset_name, "invalid"


def view_dataset(root_path: pathlib.Path) -> viewReturn:
  def get_images(dataset_path: pathlib.Path):
    # lookup 3 images.
    pics = list(files.scan_pics(dataset_path))[:1]
    return pics

  datasets = dataset.get_dataset_list(root_path)
  for dataset_name in datasets:
    res = dataset.get_dataset_path(root_path, str(dataset_name))
    pics = get_images(res)
    for pic in pics:
      yield pic, dataset_name, "_"



import glob
import pathlib
import random
import typing as t

from . import config


def read_image_area(file_path: pathlib.Path) -> config.T.Rect:
  """读取一个图片的长条矩形部分"""
  assert file_path.suffix in (".jpg", ".jpeg"), f"仅支持jpg文件, given name: {file_path}"
  assert file_path.exists(), f"{file_path}文件不存在"

  xml_filepath = config.get_xml_config_path(file_path, types="block")
  assert xml_filepath.exists(), f"{xml_filepath}文件不存在"

  return config.read_rect_from_file(xml_filepath, "block")


def get_random_image_file(root_dir: pathlib.Path) -> pathlib.Path:
  """获取随机图像文件"""
  return random.choice(list(root_dir.glob("*.jpg")))


def read_area_img(
  root: pathlib.Path,
  get_dataset: t.Callable[[], t.Union[str, pathlib.Path]],
  # range_: tuple[float, float],
  promise=False,
):
  pass


def list_images(root: pathlib.Path, dataset_name: str) -> list[pathlib.Path]:
  dataset_full_path = f"lens_6/XL/XL/{dataset_name}"
  return glob.glob(str(root / dataset_full_path / "*.jpg"))



# 统一数据集表示
# 只在一个文件里修改目录，方便快捷
# 将使用的函数放在参数上，减少依赖

from __future__ import annotations

import datetime
import os
import pathlib

# import numpy as np
import random
import typing as t

from tqdm import tqdm

from ... import files
from ... import types as T
from ...img import process
from ..join import join_with_fix, join_with_resize  # noqa
from ..view import view_dataset, view_dataset_in_rows, view_dataset_on_disk  # noqa


class Dataset(object):
  def generate_func(self) -> t.List[t.Callable[[pathlib.Path], T.ImgDataset]]:
    """生成一组用于加载数据集的函数

    Returns:
        返回一个函数列表,每个函数用于加载指定路径下的训练或测试数据
    """
    funcs = []
    names: t.List[str] = [T.x_name, T.y_name, T.x_test, T.y_test]
    for n in names:

      def get_func(path: pathlib.Path, name: str = n):
        return files.load_from_disk(path / name)

      funcs.append(get_func)

    return funcs

  def get_details(
    self,
    path: pathlib.Path,
    x: T.ImgDataset,
    y: T.LabelData,
  ) -> t.Dict:
    """获取数据集的详细信息

    Args:
        path: 数据集路径
        x: 图像数据集
        y: 标签数据

    Returns:
        包含数据集详细信息的字典,包括路径、创建时间、数据形状等
    """
    data = {}

    def create_sub(name):
      data[name] = {}

    for name in ["Dataset", "Meta"]:
      create_sub(name)

    data["Dataset"]["path"] = str(path)
    data["Meta"]["config.create_time"] = datetime.datetime.now()
    data["Dataset"]["created_time"] = ""
    data["Dataset"]["updated_time"] = ""
    data["Dataset"]["x_shape"] = x.shape
    data["Dataset"]["y_shape"] = y.shape
    return data

  def show_details(
    self,
    get_x_train: t.Callable,
    get_y_train: t.Callable,
    get_details: t.Callable[[t.Any, t.Any], t.Dict],
    write_to_file: t.Callable[[t.Dict], None],
  ):
    x, y = get_x_train(), get_y_train()
    details = get_details(x, y)
    write_to_file(details)

  def dataset_length_list(self) -> t.List[int]:
    """返回支持的数字长度列表

    Returns:
        支持的数字位数列表[5,6,7,8]
    """
    return [5, 6, 7, 8]

  def create_str_with_fill(
    self,
    number: int,
    length: int,
    total: int,
  ) -> T.DigitStr:
    """创建填充后的数字字符串

    Args:
        number: 要转换的数字
        length: 数字的长度
        total: 需要填充到的总长度

    Returns:
        填充后的数字字符串
    """
    return self.fill_digit(process.number_to_string(number, length), total)

  def create_labels_func(
    self,
    length: int,
    total: int,
  ) -> t.Callable[[int], t.Tuple[t.List[int], t.List[T.DigitStr]]]:
    """创建标签生成函数

    Args:
        length: 数字长度
        total: 填充后的总长度

    Returns:
        返回一个函数,该函数生成指定数量的随机数字及其字符串表示
    """

    def generate_nums(train_nums: int):
      numbers = []
      str_digits = []
      for _ in range(train_nums):
        number = random.randint(0, 10**length)
        numbers.append(number)
        str_digits.append(
          self.create_str_with_fill(number, length, total),
        )
      return numbers, str_digits

    return generate_nums

  def create_dataset_func(
    self,
    check_imgs: t.Callable[[T.ImgList], None],
    total: int,
  ) -> CreateDatasetFunc:
    """创建数据集生成函数

    Args:
        check_imgs: 图像检查函数
        total: 填充后的总长度

    Returns:
        返回一个函数,用于生成包含指定数量和长度的数字图像数据集
    """

    def inner(
      self,
      length: int,
      nums: int,
      gen_block_img: GenBlockImgFunc,
    ):
      _, str_digits = self.create_labels_func(length, total)(nums)

      imgs = []
      for digit in tqdm(str_digits):
        im = gen_block_img(digit)
        imgs.append(im)

      # files.write_shape(imgs, 3)
      check_imgs(imgs)

      # automatic resize the images
      imgs = process.resize_imglist(imgs, size=[37, 297])
      return imgs, str_digits

    return inner

  def get_random_dataset(
    self,
    root: pathlib.Path,
  ) -> t.Tuple[pathlib.Path, int]:
    """随机获取一个数据集

    Args:
        root: 根目录
        get_dataset_list: 获取数据集列表的函数

    Returns:
        随机选择的数据集路径和索引
    """
    datasets = list(self.get_dataset_list(root))
    random_index = random.randint(0, len(datasets) - 1)
    return datasets[random_index], random_index

  def generate_block_img(
    self,
    the_digit: T.DigitStr,
    join_func: T.JoinFunc,
    read_rand_img: t.Callable[[int | str], T.NpImage],
  ) -> T.NpImage:
    """生成数字块图像

    Args:
        the_digit: 数字字符串
        join_func: 图像拼接函数
        read_rand_img: 读取随机图像的函数

    Returns:
        拼接后的数字块图像
    """
    img_list = []
    for digit in the_digit:
      im = read_rand_img(digit)
      img_list.append(im)
    return join_func(img_list, process.size_check)

  def get_dataset_path(
    self,
    root: pathlib.Path,
    dataset_name: str,
  ) -> pathlib.Path:
    """获取数据集路径

    Args:
        root: 根目录
        dataset_name: 数据集名称

    Returns:
        完整的数据集路径
    """
    p = root / "lens_6" / "XL" / "XL" / dataset_name
    return p

  def get_dataset_list(
    self,
    root: pathlib.Path,
    default_func: t.Callable = lambda: pathlib.Path("lens_6/XL/XL"),
  ) -> t.Iterator[pathlib.Path]:
    """获取数据集列表

    Args:
        root: 根目录
        default_func: 默认路径生成函数

    Returns:
        数据集路径的迭代器
    """
    root = root / default_func()
    for dir in os.listdir(root):
      if os.path.isdir(root / dir):
        yield pathlib.Path(dir)

  def handle_datasets(
    self, root: pathlib.Path, handle_func: t.Callable[[pathlib.Path], None]
  ):
    """handle"""
    for dataset in self.get_dataset_list(root):
      handle_func(dataset)

  def fill_digit(
    self,
    digit: T.DigitStr,
    total_length: int,
  ) -> T.DigitStr:
    """填充数字字符串到指定长度

    Args:
        digit: 原始数字字符串
        total_length: 目标长度

    Returns:
        填充后的数字字符串

    Raises:
        ValueError: 当输入字符串长度小于5或大于目标长度时
    """
    digit2 = digit.copy()
    if len(digit) < 5:
      raise ValueError(f"digit lenght must > 5, {digit}")

    if len(digit) <= total_length:
      digit2.extend(["x"] * (total_length - len(digit)))
    else:
      raise ValueError(f"digit lenght must < {total_length}, {digit}")
    return digit2


class GenBlockImgFunc(t.Protocol):
  def __call__(self, digit: T.DigitStr) -> T.NpImage: ...


class SaveDatasetFunc(t.Protocol):
  def __call__(self, imgs: T.ImgList, str_digits: t.List[T.DigitStr]) -> None: ...


class CreateDatasetFunc(t.Protocol):
  def __call__(
    self,
    length: int,
    nums: int,
    gen_block_img: GenBlockImgFunc,
  ) -> t.Tuple[t.List[T.NpImage], t.List[T.DigitStr]]: ...
